'''
@file: __init__.py
@author: Jaume Bonet / Patrick Gohl
@mail:   jaume.bonet@gmail.com / patrick.gohl@upf.edu
@date:   2013
@ [oliva's lab](http://sbi.imim.es)
@class: Parameters
'''
import sys
import os
import traceback
import datetime
import time
import logging

__version__ = '0.3.2'


class Parameters(object):
    '''
    Designed to work through all the {SBI} library.
    It contains several parameters that will control (a) the amount of data
    shown to the user during the execution of {SBI} subroutines and (b) the
    file overwrite settings in a execution. See {SBI.beans.File} for more info
    in that topic.
    '''
    _LOGNAME = 'SBILOG'

    def __init__(self):
        self._verbose  = False      # > Minimum level of progress info.
                                    # Mainly, to inform of advance.
        self._debug    = False      # > Medium level of progress info.
                                    # Use to better understand the process
                                    # behind a particular function.
        self._ddebug   = False      # > Maximum level of progress info.
                                    # To check on processes that are going
                                    # to be highly repeated.
        self._overwrte = False      # > This will set the general overwrite
                                    # setting for all the execution.
                                    # It is OVER-RULED by local overwrite
                                    # parameters.
        self._warning  = True       # > Prompt warning info.
        self._error    = True       # > Prompt error info.

        self._stdout   = False      # STDOUT status

        self._fd       = logging.getLogger()  # > All messages can be redirected to
                                              # some file-handle different than STDERR
        self._fd.setLevel(logging.NOTSET)

    ##############
    # ATTRIBUTES #
    ##############

    @property
    def verbose(self):
        '''
        Minimum level of progress info.
        Mainly, to inform of advance.
        @return:  {Boolean}
        @default: _False_
        '''
        return self._verbose

    @verbose.setter
    def verbose(self, value):
        '''
        @param: {Boolean}
        '''
        self._verbose = value

    @property
    def debug(self):
        '''
        Medium level of progress info.
        Use to better understand the process behind a particular function.
        @return:  {Boolean}
        @default: _False_
        '''
        return self._debug

    @debug.setter
    def debug(self, value):
        '''
        @param: {Boolean}
        '''
        if value:
            self._verbose, self._debug = value, value
        else:
            self._debug                = value

    @property
    def deepdebug(self):
        '''
        Maximum level of progress info.
        To check on processes that are going to be highly repeated.
        @return:  {Boolean}
        @default: _False_
        '''
        return self._ddebug

    @deepdebug.setter
    def deepdebug(self, value):
        '''
        If deepdebug is set to _True_, it also sets _debug_ and _verbose_
        to _True_.
        @param: {Boolean}
        '''
        if value:
            self._ddebug = value
            self.debug   = value
        else:
            self._ddebug = value

    @property
    def warnings(self):
        '''
        Prompts warning info in places where something is not completely wrong
        but there is no need to stop the execution.
        @return:  {Boolean}
        @default: _True_
        '''
        return self._warning

    @warnings.setter
    def warnings(self, value):
        '''
        @param: {Boolean}
        '''
        self._warning = value

    @property
    def errors(self):
        '''
        Prompts error info in places where something goes wrong.
        @return:  {Boolean}
        @default: _True_
        '''
        return self._error

    @errors.setter
    def errors(self, value):
        '''
        @param: {Boolean}
        '''
        self._error = value

    @property
    def overwrite(self):
        '''
        General overwrite setting for all the execution.
        It is OVER-RULED by local overwrite
        @return:  {Boolean}
        @default: _False_
        '''
        return self._overwrte

    @overwrite.setter
    def overwrite(self, value):
        '''
        @param: {Boolean}
        '''
        self._overwrte = value

    @property
    def stdout(self):
        '''
        STDOUT output status
        @return:  {Boolean}
        @default: _False_
        '''
        return self._overwrte

    @stdout.setter
    def stdout(self, value):
        '''
        @param: {Boolean}
        '''
        self._stdout = value
        if value:
            self.log_file()

    ###########
    # METHODS #
    ###########
    def log_file(self, log_file = None):
        '''
        Manually select a file to print the output information.
        The resultant log-file is independent from the overwrite parameter.
        It opens as any regular python file (that is, it does overwrite).
        @param:    log_file
        @pdef:     Name of the log file
        @pdefault: _None_, defaults to STDERR
        @ptype:    {String}
        '''
        frmt = logging.Formatter('[{0}] '.format(self._LOGNAME) +
                                 '%(asctime)s - %(levelname)-7.7s - %(message)s',
                                 '%Y-%m-%d %H:%M')
        if log_file is None:
            handler = logging.StreamHandler()
        else:
            handler = logging.FileHandler(log_file)
        handler.setFormatter(frmt)
        self._fd.addHandler(handler)
        if log_file is not None:
            self._fd.info('[SBIglobals]: Logfile {0} created.'.format(log_file))

    def set_file(self, work_dir = os.getcwd()):
        '''
        Automatically create a file to print the output information.
        The log-file consists in the name of the current executing process,
        the process-identifier (pid) number and the .log extension.
        The resultant log-file is independent from the overwrite parameter.
        It opens as any regular python file (that is, it does overwrite).
        @param:    work_dir
        @pdef:     Name of the directory into which create the file.
        @pdefault: Current working directory.
        @ptype:    {String}
        '''
        script_name = os.path.split(os.path.splitext(sys.argv[0])[0])[1]
        log_file    = ".".join([script_name, str(os.getpid()), 'log'])
        self.log_file(os.path.join(work_dir, log_file))

    def decide_overwrite(self, local_overwrite):
        '''
        When local file overwrite has not been set (is None), the *Parameters*
        _overwrite_ attribute is called. Otherwise local overwrite
        specification has priority.
        @param: local_overwrite
        @pdef:  Selected status to control the overwrite of an already existing
                file.
        @ptype: {Boolean}
        @return: {Boolean}
        '''
        return self.overwrite if local_overwrite is None else local_overwrite

    def alert(self, level = 'verbose', source_object = None, message = None):
        '''
        Throws a message for the user.
        @param:    level
        @pdef:     specifies the minimum active level to actually show the
                   message.
        @pdefault: verbose
        @poptions: 'verbose', 'debug', 'deepdebug'
        @ptype:    {String}
        @param:    source_object
        @pdef:     object that is currently throwing a message.
        @pdefault: {None}
        @ptype:    {object}
        param:     message
        @pdef:     specific message sent by the user.
        @pdefault: {None}
        @ptype:    {String}
        '''
        if not self._active_level(level):
            return
        if source_object is None and message is None:
            return

        name = self._source_name(source_object)
        if message is None:
            message = ''
        if isinstance(message, list):
            for line in message:
                self.alert(level, source_object, line)
        else:
            if level == 'verbose':
                self._fd.info('{0}{1}'.format(name, message))
            else:
                self._fd.debug('{0}{1}'.format(name, message))

    def warn(self, source_object = None, message = None):
        '''
        Throw a warning message. Call when something should be said to the user
        about what's going on.
        @param:    source_object
        @pdef:     object that is currently throwing a warning.
        @pdefault: {None}
        @ptype:    {object}
        @param:    message
        @pdef:     specific message sent by the user.
        @pdefault: {None}
        @ptype:    {String}
        '''
        if self.warnings is False:
            return

        name = self._source_name(source_object)
        if isinstance(message, list):
            for line in message:
                self.warn(source_object, line)
        else:
            self._fd.warning('{0}{1}'.format(name, message))

    def throw(self, source_object = None, message = None,
                    error         = None, killit  = True):
        '''
        Throw an error. Call it when an error occur.
        If the _error_ attribute is _False_, it skips the effect, unless
        _killit_ is _True_, which will imply that we can not proceed given the
        error
        @param:    source_object
        @pdef:     object that is currently reporting an error.
        @pdefault: {None}
        @ptype:    {object}
        @param:    message
        @pdef:     specific message sent by the user.
        @pdefault: {None}
        @ptype:    {String}
        @param:    error
        @pdef:     {Exception} that has been captured with the error.
        @pdefault: {None}
        @ptype:    {Exception}
        @param:    killit
        @pdef:     order to stop the execution after the error has been
                   reported.
        @pdefault: _True_
        @ptype:    {Boolean}
        '''
        if self.errors is False and killit is False:
            return

        name = self._source_name(source_object)

        self._fd.error('{0}{1}'.format(name, message))

        if error is not None:
            self._fd.error(error.message)
            self._fd.error(traceback.format_exc())

        if killit:
            sys.exit()

    def success(self):
        '''
        Tells that the program has successfully finished
        '''
        self._fd.info('[SUCCESS!!]: -- Program ended as expected.')
        sys.exit(0)

    def countdown(self, max_time):
        '''
        Prints a countdown in place.
        Put it in a loop if you are waiting for something.
        @param:    max_time
        @pdef:     time to wait. in seconds.
        @ptype:    {integer}
        '''
        t  = str(datetime.timedelta(seconds=max_time))
        n  = time.localtime()
        s1 = 'Waiting for: {0} hours'.format(t)
        s2 = 'Wait started at {0}'.format(time.strftime('%X', n))
        s3 = 'on {0}'.format(time.strftime('%Y-%m-%d', n))
        sys.stderr.write('{0}\t{1} {2}\n\n'.format(s1, s2, s3))
        while max_time > 0:
            t = str(datetime.timedelta(seconds=max_time))
            sys.stderr.write('Remaining: {0} hours'.format(t))
            time.sleep(1)
            max_time -= 1
            if bool(max_time):
                sys.stderr.write('\r')
            else:
                sys.stderr.write('\r')
                t = str(datetime.timedelta(seconds=max_time))
                sys.stderr.write('Remaining: {0} hours'.format(t))
                sys.stderr.write('\n')

    ###################
    # PRIVATE METHODS #
    ###################
    def _active_level(self, level):
        '''
        Decide if a given level is active or not.
        @param:    level
        @pdef:     level to check if it is active.
        @pdefault: {None}
        @ptype:    {String}
        @return: {Boolean}
        '''
        return ((level == 'verbose'   and self.verbose) or
                (level  == 'debug'     and self.debug)   or
                (level  == 'deepdebug' and self.deepdebug))

    def _source_name(self, source_object):
        '''
        Format the name of the object calling SBIglobals
        @param:    source_object
        @pdef:     object
        @ptype:    {object}
        '''
        if isinstance(source_object, basestring):
            return '[' + source_object.upper() + ']: '
        elif source_object is not None:
            return '[' + source_object.__class__.__name__.upper() + ']: '
        else:
            return ''

SBIglobals = Parameters()
'''
The {SBIglobals} variable is used through all the library, being the one that
actually sets the {Parameters} around the different classes and functions
of {SBI}.
'''



from setuptools import setup, find_packages

with open('README.md') as f:
    readme = f.read()


setup(
    name='SBILib',
    version='0.3.2',
    license='MIT',
    author="Patrick Gohl",
    author_email='patrick.gohl@upf.edu',
    packages=find_packages(where='src'),
    package_dir={'': 'src'},
    include_package_data=True,
    long_description=readme,
    long_description_content_type="text/markdown",
    url='https://github.com/structuralbioinformatics/SBI',
    keywords='Structural Bioinformatics, Loops',
    install_requires=[
          'numpy','pandas','pynion','Requests','scipy','six'
      ],

)







# -*- coding: utf-8 -*-
"""
.. codeauthor:: Jaume Bonet <jaume.bonet@gmail.com>

.. affiliation::
    Structural BioInformatics Lab <sbi.upf.edu>
    Baldo Oliva <baldo.oliva@upf.edu>
"""
# Standard Libraries
import copy

# External Libraries
import pandas as pd

# This Library
import SBI.core as core

__all__ = ['alphabet']


class Alphabet( pd.DataFrame ):

    _aminoacids_main3    = []
    _aminoacids_extr3    = []
    _aminoacids_main1    = []
    _aminoacids_3to1     = {}
    _aminoacids_1to3     = {}
    _protein_backbone    = ['CA', 'N', 'O', 'C']

    _nucleotide_main     = []
    _nucleotide_backbone = ['P',
                            'O1P', 'O2P', 'O3P', "O5'", "C5'", "C4'", "O4'", "C3'", "O3'", "C2'", "O2'", "C1'",
                            'OP1', 'OP2', 'OP3', "O5*", "C5*", "C4*", "O4*", "C3*", "O3*", "C2*", "O2*", "C1*"]

    def __init__( self, *args, **kwargs ):
        super(Alphabet, self).__init__(*args, **kwargs)
        if self.shape[0] > 0:
            self._init_values()

    @property
    def protein_backbone( self ):
        return self._protein_backbone

    @property
    def aminoacids_main3( self ):
        return copy.copy(self._aminoacids_main3)

    @property
    def aminoacids_main1( self ):
        return copy.copy(self._aminoacids_main1)

    @property
    def nucleotide_main( self ):
        return copy.copy(self._nucleotide_main)

    @property
    def nucleotide_backbone( self ):
        return self._nucleotide_backbone

    def is_aminoacid( self, code3 ):
        if code3 in self._aminoacids_main3 or code3 in self._aminoacids_extr3:
            return True
        data = self[(self['three_letter_code'] == code3) &
                    (self['type'].str.contains('PEPTIDE LINKING'))
                    ]
        if data.shape[0] == 0:
            return False
        elif data.shape[0] == 1:
            data = data.iloc[0]
        if data['one_letter_code'] != ' ' or data['mon_nstd_parent_comp_id'] != ' ':
            self._aminoacids_extr3.append(code3)
            return True
        else:
            return False

    def is_nucleotide( self, code3 ):
        # @TODO: Improve nucleotide detection of non-main types
        # @BODY: Do a similar approach as in the aminoacid.
        if code3 in self._nucleotide_main:
            return True
        return False

    def aminoacids3to1( self, code3 ):
        try:
            return self._aminoacids_3to1[code3]
        except KeyError:
            data = self[(self['three_letter_code'] == code3) &
                        (self['type'].str.contains('PEPTIDE LINKING'))
                        ]
            if data.shape[0] == 0:
                if core.get_option('data', 'strict'):
                    raise Unknown3LetterCodeError('Unknown code {}'.format(code3))
                else:
                    self._aminoacids_3to1[code3] = 'X'
            elif data.shape[0] > 1:
                raise Ambiguous3LetterCodeError('Multiple hits for {}'.format(code3))
            else:
                data = data.squeeze()
                if data['one_letter_code'] != ' ':
                    if len(data['one_letter_code']) == 1:
                        self._aminoacids_3to1[code3] = data['one_letter_code']
                    else:
                        self._aminoacids_3to1[code3] = '[' + data['one_letter_code'] + ']'
                elif data['mon_nstd_parent_comp_id'] != ' ':
                    self._aminoacids_3to1[code3] = self.aminoacids3to1(data['mon_nstd_parent_comp_id'])
                else:
                    if core.get_option('data', 'strict'):
                        raise Unknown3LetterCodeError('Unknown code {}'.format(code3))
                    else:
                        self._aminoacids_3to1[code3] = 'X'
            return self._aminoacids_3to1[code3]

    def aminoacids1to3( self, code1 ):
        try:
            return self._aminoacids_1to3[code1]
        except KeyError:
            data = self[(self['one_letter_code'] == code1) &
                        (self['type'].str.contains('PEPTIDE LINKING'))
                        ]
            if data.shape[0] == 0:
                if core.get_option('data', 'strict'):
                    raise Unknown1LetterCodeError('Unknown code {}'.format(code1))
                else:
                    self._aminoacids_1to3[code1] = 'UNK'
            else:
                self._aminoacids_1to3[code1] = data.iloc[0]['three_letter_code']
            return self._aminoacids_1to3[code1]

    def _init_values( self ):
        aas = self[((self['type'].str.contains('L-PEPTIDE LINKING'))  | (self['type'] == 'PEPTIDE LINKING')) &
                   (self['pdbx_initial_date'] == '1999-07-08') &
                   (self['mon_nstd_parent_comp_id'] == ' ') &
                   (~self['one_letter_code'].isin(['X', ' ']))
                   ]
        self._aminoacids_main1 = list(aas['one_letter_code'])
        self._aminoacids_main3 = list(aas['three_letter_code'])
        self._aminoacids_3to1 = pd.Series(aas.one_letter_code.values, index=aas.three_letter_code).to_dict()
        self._aminoacids_1to3 = pd.Series(aas.three_letter_code.values, index=aas.one_letter_code).to_dict()
        self._aminoacids_1to3['X'] = 'UNK'

        nns = self[(self['type'].str.contains('NA LINKING')) & (self['mon_nstd_parent_comp_id'] == ' ') &
                   (self['pdbx_type'] == 'ATOMN') & (self['name'].str.isupper()) &
                   ((self['three_letter_code'].str.len() == 1) | ((self['three_letter_code'].str.len() == 2) &
                    (self['three_letter_code'].str.startswith('D')))) ]

        self._nucleotide_main = list(nns['three_letter_code'])


class Unknown3LetterCodeError( KeyError ):
    """Raised when a new 3to1 code translation is requested on an unknown 3 letter code.
    """


class Unknown1LetterCodeError( KeyError ):
    """Raised when a new 1to3 code translation is requested on an unknown 1 letter code.
    """


class Ambiguous3LetterCodeError( KeyError ):
    """Raised when a 3 letter code generates more than one hit.
    """


alphabet = Alphabet(pd.read_csv(core.get_option('data', 'alphabet')))
"""
:data:`.alphabet` is the main container to hold static data necessary to interpret PDB codes for compound.

The data contained in this variable is loaded at runtime and generated from the
`PDBeChem <ftp://ftp.ebi.ac.uk/pub/databases/msd/pdbechem/>`_ database.
"""



import json

from SBILib.beans    import JSONer
from SBILib.sequence import Sequence


class Uniprot(JSONer):

    def __init__(self, entry_name, status):
        '''
        @param:    entry_name
        @pdef:     entry name of the new uniprot entry
        @ptype:    {String}

        @param:    status
        @pdef:     To distinguish the fully annotated entries in the Swiss-Prot
                   section of the UniProt Knowledgebase from the computer-annotated
                   entries in the TrEMBL section
        @ptype:    {String}
        '''
        self._entry     = entry_name
        self._status    = status
        self._version   = 0
        self._accession = []
        self._taxid     = None
        self._hostid    = []
        self._referals  = {}
        self._sequence  = ''

    ##############
    # ATTRIBUTES #
    ##############
    @property
    def entry_name(self):
        '''
        Uniprot unique identifier

        @return: {String}
        '''
        return self._entry

    @property
    def section(self):
        '''
        To which section the entry belongs.

        @return: {String}
        '''
        return 'swissprot' if self.is_swissprot else 'trembl'

    @property
    def version(self):
        '''
        actual version of the entry

        @return: {Integer}
        '''
        return self._version

    @version.setter
    def version(self, value):
        '''
        @param:    value
        @pdef:     version of the entry
        @ptype:    {Integer}
        '''
        self._version = int(value)

    @property
    def accession(self):
        '''
        List of accession numbers for the entry

        @return: {List}
        '''
        return self._accession

    @accession.setter
    def accession(self, value):
        '''
        @param:    value
        @pdef:     list of accession numbers
        @ptype:    {List}
        '''
        self._accession.extend(value)

    @property
    def taxid(self):
        '''
        Specie to which the protein belongs.

        @return: {String}
        '''
        return self._taxid

    @taxid.setter
    def taxid(self, value):
        '''
        @param:    value
        @pdef:     Specie to which the protein belongs.
        @ptype:    {Integer}
        '''
        self._taxid = int(value)

    @property
    def host(self):
        '''
        List of species to which the protein specie infects.

        @return: {List}
        '''
        return self._hostid

    @host.setter
    def host(self, value):
        '''
        @param:    value
        @pdef:     Specie to which the protein specie infects.
        @ptype:    {Integer}
        '''
        self._hostid.append(int(value))

    @property
    def referals(self):
        '''
        References from other databases.
        Altough being all together, each reference has its specificities.
        Check them in the [Uniprot manual](http://web.expasy.org/docs/userman.html#DR_line)

        @return: {List} of {Dictionary}
        '''
        return self._referals

    @referals.setter
    def referals(self, value):
        '''
        @param:    value
        @pdef:     database reference info
        @ptype:    {List}
        '''
        db = value.pop(0)
        self._referals.setdefault(db, [])
        reference = {}
        reference['id'] = value.pop(0)
        if bool(len(value)):
            opt1 = value.pop(0).strip()
            if db == 'RefSeq':
                reference['nuc_seq_id'] = opt1
            elif db in ['InterPro', 'PANTHER', 'Pfam', 'PIR', 'PRINTS', 'ProDom',
                        'REBASE', 'SMART', 'SUPFAM', 'TIGRFAMs', 'PeroxiBase',
                        'PROSITE']:
                reference['name'] = opt1
            elif db == 'PDB':
                reference['method'] = opt1
            elif db in ['dictyBase', 'EcoGene', 'FlyBase', 'CGD', 'HGNC', 'MGI',
                        'PATRIC', 'RGD', 'SGD', 'Xenbase', 'ZFIN']:
                reference['gene_designation'] = opt1
            elif db == 'GO':
                reference['ontology'] = opt1.split(':')[0]
            elif db in ['HAMAP', 'PIRSF', 'CAZy', 'GeneFarm', 'TCDB']:
                reference['family'] = opt1
            elif db in ['WormBase', 'Ensembl', 'EnsemblBacteria', 'EnsemblFungi',
                        'EnsemblMetazoa', 'EnsemblPlants', 'EnsemblProtists',
                        'EMBL']:
                reference['protein_id'] = opt1
            elif db == 'VectorBase':
                reference['species_of_origin'] = opt1
            elif db in ['IntAct', 'BioGrid']:
                reference['num_interactors'] = opt1
            elif db == 'SMR':
                reference['range'] = opt1
            elif db == 'MIM':
                reference['type'] = opt1
            elif db == 'Allergome':
                reference['allergen'] = opt1
            elif db in ['ArachnoServer', 'ConoServer']:
                reference['toxin'] = opt1
            elif db == 'Orphanet':
                reference['disease'] = opt1
            elif db in ['ChiTaRS', 'UCSC', 'BRENDA']:
                reference['organism'] = opt1
            elif db == 'UniPathway':
                reference['reaction'] = opt1
            elif db == 'DrugBank':
                reference['drug'] = opt1
            if bool(len(value)):
                opt2 = value.pop(0).strip()
                if db in ['Gene3D', 'HAMAP', 'PANTHER', 'Pfam', 'PIRSF', 'ProDom',
                          'SMART', 'SUPFAM', 'TIGRFAMs', 'PROSITE']:
                    reference['hits'] = opt2.rstrip('.')
                elif db in ['WormBase', 'Ensembl', 'EnsemblBacteria', 'EnsemblFungi',
                            'EnsemblMetazoa', 'EnsemblPlants', 'EnsemblProtists']:
                    reference['gene_id'] = opt2
                elif db == 'GO':
                    reference['evidence'] = opt2.split(':')[0]
                elif db == 'PDB':
                    reference['resolution'] = opt2
                elif db == 'EMBL':
                    reference['status'] = opt2
                if bool(len(value)):
                    opt3 = value.pop(0).strip()
                    if db == 'PDB':
                        reference['ranges'] = opt3
                    elif db == 'WormBase':
                        reference['gene_designation'] = opt3
                    elif db == 'EMBL':
                        reference['molecule_type'] = opt3
        self._referals[db].append(reference)

    @property
    def sequence(self):
        '''
        Protein sequence

        @return: {Sequence}
        '''
        return Sequence(self.entry_name, self._sequence)

    @sequence.setter
    def sequence(self, value):
        '''
        @param:    value
        @pdef:     section of the protein sequence
        @ptype:    {String}
        '''
        self._sequence += value

    ############
    # BOOLEANS #
    ############
    @property
    def is_swissprot(self):
        '''
        Checks if the entry belongs to Swiss-Prot

        @return: {Boolean}
        '''
        return self._status == 'Reviewed'

    @property
    def is_trembl(self):
        '''
        Checks if the entry belongs to TrEMBL

        @return: {Boolean}
        '''
        return self._status == 'Unreviewed'

    def has_reference_to(self, external_db):
        '''
        Check if the protein has a reference to a given database.

        @param:    external_db
        @pdef:     identifier of the external database
        @ptype:    {String}

        @return: {Boolean}
        '''
        return external_db in self.referals

    ###########
    # METHODS #
    ###########
    @staticmethod
    def grab(json_line):
        '''
        Retrieve the object data from a json line.

        @param:    json_line
        @pdef:     line of data.
        @ptype:    {String} or {Dictionary}

        @return: {Uniprot}
        '''
        if isinstance(json_line, basestring):
            json_line = json.loads(json_line.strip())

        data = Uniprot('', '')
        data.__dict__ = json_line

        return data

    def as_dict(self):
        return self.__dict__



class CDhit(object):
    def __init__(self, clusterid):
        self._id        = clusterid
        self._master    = None
        self._sequences = {}

    @property
    def identifier(self): return self._id
    @property
    def master(self):     return self._master
    @property
    def sequences(self):  return self._sequences

    def is_master(self, seq):
        return self._master.name == seq

    def is_sequence(self, seq):
        return seq in self._sequences

    def add_sequence(self, cdhithomolog):
        if cdhithomolog.is_master:
            self._master = cdhithomolog
        else:
            self._sequences[cdhithomolog.name] = cdhithomolog

    def __repr__(self):
        text = []
        text.append('Cluster {0.identifier}:'.format(self))
        text.append('\tMaster Sequence: {0.master}'.format(self))
        for s in self.sequences:
            text.append('\t\t{0}'.format(self.sequences[s]))
        return '\n'.join(text)





# -*- coding: utf-8 -*-
"""
.. codeauthor:: Jaume Bonet <jaume.bonet@gmail.com>

.. affiliation::
    Structural BioInformatics Lab <sbi.upf.edu>
    Baldo Oliva <baldo.oliva@upf.edu>
"""
# Standard Libraries

# External Libraries
import pandas as pd

# This Library
import SBI.core as core

__all__ = ['properties']


class Properties( pd.DataFrame ):
    def get_surface( self, residue_type ):
        """Get the expected **surface** for a query amino acid.

        .. note::
            Depends on global configuration option ``data.surface`` to determine
            the source from which the surface data is obtained. There are 4 main
            data sources for surface information:

            * _tien2013t_ and _tien2013e_: Experimental (e) and theoretical (t) surface \
                data provided by Tien, M.Z. _et al._ (2013). [Maximum allowed solvent accessibilites \
                of residues in proteins.](https://doi.org/10.1371%2Fjournal.pone.0080635) **PLoS ONE**.
            * _miller87_: Surface data obtained from Miller, S. _et al._ (1987). [Interior and surface \
                of monomeric proteins.](https://doi.org/10.1016%2F0022-2836%2887%2990038-6) **JMB**.
            * _rose85_: Surface data obtained from Rose, G.D. _et al._ (1985). [Hydrophobicity of amino \
                acid residues in globular proteins.](https://doi.org/10.1126%2Fscience.4023714) **Science**.

        :param str residue_type: Amino acid type (1 or 3 letter code).

        :returns: :class:`.float`

        :raises:
            :IndexError: If no surface can be found for the specified amino acid code
                (i.e. if the code is not found)
        """
        surf_source = core.get_option('data', 'surface')
        residue_type = residue_type.strip()
        col = 'one_letter_code' if len(residue_type) == 1 else 'three_letter_code'
        return self[self[('residue', col)] == residue_type][('surface', surf_source)].values[0]


colnames = pd.MultiIndex.from_tuples([
    ('residue', 'one_letter_code'), ('residue', 'three_letter_code'),
    ('surface', 'tien2013t'), ('surface', 'tien2013e'), ('surface', 'miller87'), ('surface', 'rose85')
])
properties = Properties([
    ['A', 'ALA', 129.0, 121.0, 113.0, 118.1],
    ['R', 'ARG', 274.0, 265.0, 241.0, 256.0],
    ['N', 'ASN', 195.0, 187.0, 158.0, 165.5],
    ['D', 'ASP', 193.0, 187.0, 151.0, 158.7],
    ['C', 'CYS', 167.0, 148.0, 140.0, 146.1],
    ['E', 'GLU', 223.0, 214.0, 183.0, 186.2],
    ['Q', 'GLN', 225.0, 214.0, 189.0, 193.2],
    ['G', 'GLY', 104.0,  97.0,  85.0,  88.1],
    ['H', 'HIS', 224.0, 216.0, 194.0, 202.5],
    ['I', 'ILE', 197.0, 195.0, 182.0, 181.0],
    ['L', 'LEU', 201.0, 191.0, 180.0, 193.1],
    ['K', 'LYS', 236.0, 230.0, 211.0, 225.8],
    ['M', 'MET', 224.0, 203.0, 204.0, 203.4],
    ['F', 'PHE', 240.0, 228.0, 218.0, 222.8],
    ['P', 'PRO', 159.0, 154.0, 143.0, 146.8],
    ['S', 'SER', 155.0, 143.0, 122.0, 129.8],
    ['T', 'THR', 172.0, 163.0, 146.0, 152.5],
    ['W', 'TRP', 285.0, 264.0, 259.0, 266.3],
    ['Y', 'TYR', 263.0, 255.0, 229.0, 236.8],
    ['V', 'VAL', 174.0, 165.0, 160.0, 164.5]
], columns=colnames)
"""
:data:`.properties` is the main container to hold static data necessary to interpret PDB codes for compound.

The data contained in this variable is loaded at runtime and generated from the
`PDBeChem <ftp://ftp.ebi.ac.uk/pub/databases/msd/pdbechem/>`_ database.
"""


import urllib.request

def download_pdb(uniprot_accession):
	url = f"https://alphafold.ebi.ac.uk/files/AF-{uniprot_accession}-F1-model_v4.pdb"
	urllib.request.urlretrieve(url, f"AlphaPrediction_{uniprot_accession}.pdb")
	return(f"AlphaPrediction_{uniprot_accession}.pdb")




"""File

author: jbonet
date:   02/2013

@oliva's lab
"""

"""
Import Standard Libraries
"""
import os
import math
import gzip
import warnings

"""
Dependences in SBI library
"""
from SBILib.error import FileError
from SBILib       import SBIglobals


class File(object):
    """The File class manages certain aspects of the IO process

        File allows to seamless control some aspects about working with files.
        Namely, it will allow to:
            - work with both regular and gzipped files.
            - quickly obtain paths, file name, extension...
            - check the viability reading/writing a given file
            - avoid/ignore overwrite an existing file
            - split a file into several

        Initializing parameters
            + file_name : [M] Name of the file
            + action    : Intention.
                          Available options are (r, rb) to read and (w, wb) to write
                          @Default: 'r'
            + overwrite : On write mode, ignores if a file with the same name existed before
                          @Default: False
            @Raises FileError

        Attributes:
            > full      : Returns a full path to the file.
            > dir       : Returns the path UP TO the file.
            > lastdir   : Returns ONLY the last dir containing the file.
            > name      : Returns ONLY the file name.
            > prefix    : Returns everything EXCEPT the last extension of the file. (exclude gzip)
            > first_prefix: Returns the absolute firts part of the file
            > extension : Returns the LAST extension of the file. (exclude gzip)
            > descriptor: Opens (first call) and returns the file-descriptor
            * action    : Defines what to do with the file.
                          Available options are listed in the available_action global variable.
                          When changing the action, the file's possibilities are rechecked
                          @Raises FileError
        Booleans:
            > is_gziped : Evaluates whether the file is/has to be compressed.
                          Evaluation based SOLELY on the '.gz' termination.

        Methods:
            - close()   : Closes the filehandle
                            - clean (bool):   deletes the file if size 0.
                                    @Default: False
                          @Raises FileError

            > split()   : Splits a given file in multiple files
                            - start_separator (string):  Line startswith. Is a line to mark the line that starts the new file
                                              @Default:  None, only other parameters are considered
                            - end_separator   (string):  Line startswith. Is a line to mark the line that ends the new file
                                              @Default:  None, only other parameters are considered
                            - size            (int):     Max allowed file size. IN BYTES
                                              @Default:  1073741824 (1GB)
                            - sections        (int):     Number of sections to split file into. Overwrites 'size'
                                              @Default:  None
                            - overwrite       (bool):    Overwrite previous files of the same name
                          @Returns an array with the File object of each created file.
                          @Raises FileError
                          @Raises AttributeError if both start_separator and end_separator are defined

        Requirements:
            * os, sys, math, gzip, warnings
            < SBI.error.FileError
            < SBI.SBIglobals

        Usage:
            from SBILib.beans.File import File
            newReadFile = File(file_name = 'test.txt', action = 'r')
            for line in newReadFile.descriptor:
                //DO STUFF
            newReadFile.close()
            newWriteFile = File(file_name = 'test.txt', action = 'w', overwrite = True)
            newWriteFile.write(//STUFF)
            newWriteFile.close()
    """
    write_action     = set(['w', 'a', 'ar', 'wb'])
    read_action      = set(['r', 'rb'])
    available_action = write_action.union(read_action)


    def __init__(self, file_name=None, action='r', overwrite=None):
        if file_name is None: raise FileError(0)    # Raise 'No file specified' Error
        self._file  = file_name

        SBIglobals.alert('debug', self, 'Preparing File: {0}'.format(self.full))

        self._action    = None
        self._check_action(action.lower())  #action must be valid

        self._fd        = None
        # Local overwrite takes precedence over Global overwrite
        self._overwrite = SBIglobals.decide_overwrite(overwrite)

        self._check_file()

    """ATTRIBUTES"""
    @property
    def full(self):        return os.path.abspath(self._file)

    @property
    def dir(self):         return os.path.split(self.full)[0]

    @property
    def lastdir(self):     return os.path.basename(self.dir)

    @property
    def name(self):        return os.path.split(self.full)[-1]

    @property
    def prefix(self):
        if self.is_gziped: return os.path.splitext(os.path.splitext(self.name)[0])[0]
        else:              return os.path.splitext(self.name)[0]

    @property
    def first_prefix(self):return self.name.split('.')[0]

    @property
    def extension(self):
        if self.is_gziped: return os.path.splitext(os.path.splitext(self.name)[-1])[-1]
        else:              return os.path.splitext(self.name)[-1]

    @property
    def action(self):      return self._action
    @action.setter
    def action(self, value):
        self._check_action(value.lower())
        self._check_file()

    @property
    def descriptor(self):
        if self._fd == None:    # If it was not open, it needs to be opened
            if not self.is_gziped: self._fd = open(self.full, self.action)
            else:                  self._fd = gzip.open(self.full, self.action)
        return self._fd

    """BOOLEANS"""
    @property
    def is_gziped(self): return os.path.splitext(self.name)[-1] == '.gz'


    def close(self, clean = False):
        if self._fd is not None:    # only close something previously open
            self._fd.close()
            self._fd = None
            if clean and os.path.getsize(self.full) == 0:   # Delete empty files
                warnings.warn("The output file %s is empty and it's going to be deleted" %self.full)
                os.unlink(self.full)
        # else: raise FileError(2, self.full) # Why would you close something already closed?

    def write(self, line):
        if not self._action in self.write_action:
            raise FileError(6)
        if self._fd is None:
            if not self.is_gziped: self._fd = open(self.full, self.action)
            else:                  self._fd = gzip.open(self.full, self.action)
        self._fd.write(line)

    def split(self, start_separator = None, end_separator = None, size = 1073741824, sections = None, overwrite = None):
        """
            > split()   : Splits a given file in multiple files
                            - start_separator (string):  Line startswith. Is a line to mark the line that starts the new file
                                              @Default:  None, only other parameters are considered
                            - end_separator   (string):  Line startswith. Is a line to mark the line that ends the new file
                                              @Default:  None, only other parameters are considered
                            - size            (int):     Max allowed file size. IN BYTES
                                              @Default:  1073741824 (1GB)
                            - sections        (int):     Number of sections to split file into. Overwrites 'size'
                                              @Default:  None
                            - overwrite       (bool):    Overwrite previous files of the same name
                          @Returns an array with the File object of each created file.
                          @Raises FileError
                          @Raises AttributeError if both start_separator and end_separator are defined
        """
        # A defined separator will avoid splitting data
        # For example, uniprot files separate between proteins with '//', then by declaring the separator
        # we can avoid a protein from being split between two files.

        counter  = 1
        newfiles = []

        if (size is None and sections is None) or not isinstance(size, int):
            raise FileError(5, self.__name__, 'split')
        if (start_separator is not None and end_separator is not None):
            raise AttributeError('Both start and end separators can not be defined simultaniously\n')

        # section TAKES PRECEDENCE over size...
        if sections is not None: size     = math.ceil(os.path.getsize(self.full) / float(sections))
        else:                    sections = math.ceil(os.path.getsize(self.full) / size)

        # preparing partition file names
        outputfile = os.path.join(self.dir, self.prefix + '.{0:003d}' + self.extension)
        if self.is_gziped: outputfile += '.gz'

        # Local overwrite takes precedence over Global overwrite
        overwrite = SBIglobals.decide_overwrite(overwrite)

        SBIglobals.alert('verbose', self, 'Dividing {0.full} into {1} files (aprox.)'.format(self, sections))

        newfiles.append(File(file_name = outputfile.format(counter), action = 'w', overwrite = overwrite))
        for line in self.descriptor:
            if (start_separator is None or line.startswith(start_separator)) and os.path.getsize(newfiles[-1].full) >= size:
                newfiles[-1].close()
                counter += 1
                newfiles.append(File(file_name = outputfile.format(counter), action = 'w', overwrite = overwrite))

            newfiles[-1].write(line)

            if (end_separator is None or line.startswith(end_separator)) and os.path.getsize(newfiles[-1].full) >= size:
                newfiles[-1].close()
                counter += 1
                newfiles.append(File(file_name = outputfile.format(counter), action = 'w', overwrite = overwrite))
        newfiles[-1].close()
        self.close()
        return newfiles

    """PRIVATE METHODS"""
    def _check_action(self, action):
        if not action in self.available_action:  # Raise 'Wrong action' Error
            raise File(1, action, self.available_action)
        if self.is_gziped and not action.endswith('b'): action += 'b'   # Better if working with compressed files

        self._action = action

        SBIglobals.alert('debug', self, '\tAction {0} is OK...'.format(self._action))

    def _check_file(self):
        if self._action.startswith('r'):
            if not os.path.isfile(self.full):     raise FileError(3, self.full, 'noexists')
            if not os.access(self.full, os.R_OK): raise FileError(4, self.full, 'read')
        if self._action.startswith('w') or self._action.startswith('a'):
            if os.path.isfile(self.full):
                if not self._overwrite:           raise FileError(3, self.full, 'exists')
            if not os.path.isdir(self.dir):       raise FileError(4, self.dir,  'nodir')
            if not os.access(self.dir, os.W_OK):  raise FileError(4, self.dir,  'write')

        SBIglobals.alert('debug', self, '\tFile is OK...')

    """OVERWRITE INHERITED FUNCTIONS"""
    def __repr__(self):
        return "[{0.__class__}]: {0.full}".format(self)
